/* Pango
 * pangogp.c: Routines for handling FreeType2 fonts with 'GNOME Print' library
 *
 * Copyright (C) 1999 Red Hat Software
 * Copyright (C) 2000 Tor Lillqvist
 * Copyright (C) 2001, 2002 Sun Microsystems
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <glib.h>

#include <ft2build.h>
#include FT_FREETYPE_H

#include "pango-utils.h"
#include "pangogp.h"
#include "pangogp-private.h"
#include "pangogp-context.h"
#include "pango-attributes-private.h"
#include "pango-layout-private.h"
#include "pango-context-private.h"

#define PANGO_TYPE_GP_FONT              (pango_gp_font_get_type ())
#define PANGO_GP_FONT(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_GP_FONT, PangoGPFont))
#define PANGO_GP_FONT_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), PANGO_TYPE_GP_FONT, PangoGPFontClass))
#define PANGO_GP_IS_FONT(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_GP_FONT))
#define PANGO_GP_IS_FONT_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), PANGO_TYPE_GP_FONT))
#define PANGO_GP_FONT_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), PANGO_TYPE_GP_FONT, PangoGPFontClass))

typedef struct _PangoGPFontClass   PangoGPFontClass;
typedef struct _PangoGPMetricsInfo PangoGPMetricsInfo;

struct _PangoGPFontClass
{
  PangoFontClass parent_class;
};

struct _PangoGPMetricsInfo
{
  const char       *sample_str;
  PangoFontMetrics *metrics;
};

static PangoFontClass *parent_class;	/* Parent class structure for PangoGPFont */

static void pango_gp_font_class_init (PangoGPFontClass *class);
static void pango_gp_font_init       (PangoGPFont      *gpfont);
static void pango_gp_font_dispose    (GObject         *object);
static void pango_gp_font_finalize   (GObject         *object);

static PangoFontDescription *pango_gp_font_describe          (PangoFont            *font);

static PangoEngineShape *    pango_gp_font_find_shaper       (PangoFont            *font,
							       PangoLanguage        *language,
							       guint32               ch);

static void                  pango_gp_font_get_glyph_extents (PangoFont            *font,
							       PangoGlyph            glyph,
							       PangoRectangle       *ink_rect,
							       PangoRectangle       *logical_rect);

static PangoFontMetrics *    pango_gp_font_get_metrics       (PangoFont            *font,
							       PangoLanguage        *language);
  
static GType pango_gp_font_get_type (void);

PangoGPFont *
_pango_gp_font_new (PangoFontMap    *fontmap,
		    FcPattern  *pattern)
{
  PangoGPFont *gpfont;
  double d;

  g_return_val_if_fail (fontmap != NULL, NULL);
  g_return_val_if_fail (pattern != NULL, NULL);

  gpfont = (PangoGPFont *)g_object_new (PANGO_TYPE_GP_FONT, NULL);

  gpfont->fontmap = g_object_ref (fontmap);
  gpfont->font_pattern = pattern;
  FcPatternReference (gpfont->font_pattern);

  gpfont->description = _pango_gp_font_desc_from_pattern (pattern, TRUE);
  if (FcPatternGetDouble (pattern, FC_PIXEL_SIZE, 0, &d) == FcResultMatch)
    gpfont->size = d*PANGO_SCALE;

  gpfont->face = NULL;
  gpfont->gnome_face = NULL;
  gpfont->gnome_font = NULL;

  _pango_gp_font_map_add (gpfont->fontmap, gpfont);

  return gpfont;
}

static gboolean
set_unicode_charmap (FT_Face face)
{
  int charmap;

  for (charmap = 0; charmap < face->num_charmaps; charmap++)
    if (face->charmaps[charmap]->encoding == ft_encoding_unicode)
      {
	FT_Error error = FT_Set_Charmap(face, face->charmaps[charmap]);
	return error == FT_Err_Ok;
      }

  return FALSE;
}

static void
load_fallback_face (PangoGPFont *gpfont,
		    const char   *original_file)
{
  FcPattern *sans;
  FcPattern *matched;
  FcResult result;
  FT_Error error;
  FcChar8 *filename2 = NULL;
  gchar *name;
  int id;
  
  sans = FcPatternBuild (NULL,
			 FC_FAMILY, FcTypeString, "sans",
			 FC_SIZE, FcTypeDouble, (double)pango_font_description_get_size (gpfont->description)/PANGO_SCALE,
			 NULL);
  
  matched = FcFontMatch (0, sans, &result);
  
  if (FcPatternGetString (matched, FC_FILE, 0, &filename2) != FcResultMatch)
    goto bail1;
  
  if (FcPatternGetInteger (matched, FC_INDEX, 0, &id) != FcResultMatch)
    goto bail1;
  
  error = FT_New_Face (_pango_gp_font_map_get_library (gpfont->fontmap),
		       (char *) filename2, id, &gpfont->face);
  
  
  if (error)
    {
    bail1:
      name = pango_font_description_to_string (gpfont->description);
      g_warning ("Unable to open font file %s for font %s, exiting\n", filename2, name);
      exit (1);
    }
  else
    {
      name = pango_font_description_to_string (gpfont->description);
      g_warning ("Unable to open font file %s for font %s, falling back to %s\n", original_file, name, filename2);
      g_free (name);
    }

  if (!set_unicode_charmap (gpfont->face))
    {
      g_warning ("Unable to load unicode charmap from file %s, exiting\n", filename2);
      exit (1);
    }
  
  FcPatternDestroy (sans);
  FcPatternDestroy (matched);
}

/**
 * pango_gp_font_get_face:
 * @font: a #PangoFont
 * 
 * Returns the native FreeType2 FT_Face structure used for this PangoFont.
 * This may be useful if you want to use FreeType2 functions directly.
 * 
 * Return value: a pointer to a #FT_Face structure, with the size set correctly
 **/
FT_Face
pango_gp_font_get_face (PangoFont      *font)
{
  PangoGPFont *gpfont = (PangoGPFont *)font;
  FT_Error error;
  FcPattern *pattern;
  FcChar8 *filename;
  FcBool hinting, autohint;
  int id;

  pattern = gpfont->font_pattern;

  if (!gpfont->face)
    {
      /* If we add support for not antialising, then we should
       * need to conditionalize NO_BITMAP on that.
       */
      gpfont->load_flags = FT_LOAD_NO_BITMAP;

      /* disable hinting if requested */
      if (FcPatternGetBool (pattern, FC_HINTING, 0, &hinting) != FcResultMatch)
	hinting = FcTrue;

      if (!hinting)
        gpfont->load_flags |= FT_LOAD_NO_HINTING;

      /* force autohinting if requested */
      if (FcPatternGetBool (pattern, FC_AUTOHINT, 0, &autohint) != FcResultMatch)
	autohint = FcFalse;

      if (autohint)
        gpfont->load_flags |= FT_LOAD_FORCE_AUTOHINT;

      if (FcPatternGetString (pattern, FC_FILE, 0, &filename) != FcResultMatch)
	      goto bail0;
      
      if (FcPatternGetInteger (pattern, FC_INDEX, 0, &id) != FcResultMatch)
	      goto bail0;

      error = FT_New_Face (_pango_gp_font_map_get_library (gpfont->fontmap),
			   (char *) filename,
			   id,
			   &gpfont->face);
      if (error != FT_Err_Ok)
	{
	bail0:
	  load_fallback_face (gpfont, filename);
	}

      g_assert (gpfont->face);
  
      if (!set_unicode_charmap (gpfont->face))
	{
	  g_warning ("Unable to load unicode charmap from font file %s", filename);
      
	  FT_Done_Face (gpfont->face);
	  gpfont->face = NULL;

	  load_fallback_face (gpfont, filename);
	}

      error = FT_Set_Char_Size (gpfont->face,
				PANGO_PIXELS_26_6 (gpfont->size),
				PANGO_PIXELS_26_6 (gpfont->size),
				72 * PANGO_SCALE / 64,
				72 * PANGO_SCALE / 64);

      if (error)
	g_warning ("Error in FT_Set_Char_Size: %d", error);
    }
  
  return gpfont->face;
}

static GType
pango_gp_font_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (PangoGPFontClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) pango_gp_font_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (PangoGPFont),
        0,              /* n_preallocs */
        (GInstanceInitFunc) pango_gp_font_init,
      };
      
      object_type = g_type_register_static (PANGO_TYPE_FONT,
                                            "PangoGPFont",
                                            &object_info, 0);
    }
  
  return object_type;
}

static void 
pango_gp_font_init (PangoGPFont *gpfont)
{
  gpfont->face = NULL;

  gpfont->size = 0;

  gpfont->metrics_by_lang = NULL;

  gpfont->glyph_info = g_hash_table_new (NULL, NULL);
}

static void
pango_gp_font_class_init (PangoGPFontClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);
  PangoFontClass *font_class = PANGO_FONT_CLASS (class);

  parent_class = g_type_class_peek_parent (class);
  
  object_class->finalize = pango_gp_font_finalize;
  object_class->dispose = pango_gp_font_dispose;
  
  font_class->describe = pango_gp_font_describe;
  font_class->get_coverage = pango_gp_font_get_coverage;
  font_class->find_shaper = pango_gp_font_find_shaper;
  font_class->get_glyph_extents = pango_gp_font_get_glyph_extents;
  font_class->get_metrics = pango_gp_font_get_metrics;
}

#if 0
static void
pango_gp_debug_dump_glyph (PangoFont *font,
			       PangoGlyphInfo *gi,
			       gchar     *this_char)
{
  PangoRectangle ink, logical;
  FT_Face face;

  pango_font_get_glyph_extents (font,
				gi->glyph,
				&ink,
				&logical);

  face = pango_gp_font_get_face (font);

  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "      glyph ink: x='%1.1f' y='%1.1f' w='%1.1f' h='%1.1f'",
	 (float) ink.x / PANGO_SCALE,
	 (float) ink.y / PANGO_SCALE,
	 (float) ink.width / PANGO_SCALE,
	 (float) ink.height / PANGO_SCALE);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            log: x='%1.1f' y='%1.1f' w='%1.1f' h='%1.1f'",
	 (float) logical.x / PANGO_SCALE,
	 (float) logical.y / PANGO_SCALE,
	 (float) logical.width / PANGO_SCALE,
	 (float) logical.height / PANGO_SCALE);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            off: x='%1.1f' y='%1.1f'",
	 (float) gi->geometry.x_offset,
	 (float) gi->geometry.y_offset);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            geo_w='%1.1f' fname='%s' sname='%s'",
	 (float) gi->geometry.width,
	 face->family_name,
	 face->style_name);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            units_per_EM='%d' ascender='%d' descender='%d' height='%d'",
	 face->units_per_EM,
	 face->ascender,
	 face->descender,
	 face->height);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            font='%s'",
	 pango_font_description_to_string (pango_font_describe (font)));
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            char='%04x' kern='%s'",
	 *this_char,
	 FT_HAS_KERNING(face) ? "yes" : "no");
}
#endif

/**
 * pango_gp_render:
 * @context:     the #PangoContext representing where to render.
 * @layout_text: the text of the layout.
 * @run:         the #PangoLayoutRun to render.
 * @font:        the font in which to draw the string.
 * @glyphs:      the glyph string to draw.
 * @x:           the x position of the start of the string (in #PangoGlyphUnit).
 * @y:           the y position of the baseline (in #PangoGlyphUnit).
 *
 * Renders a PangoGlyphString onto a Gp PDF document.
 **/
void
pango_gp_render (PangoContext     *context,
		 const gchar      *layout_text,
		 PangoLayoutRun   *run,
		 PangoFont        *font,
		 PangoGlyphString *glyphs,
		 gint              x, 
		 gint              y)
{
  /*PangoLayoutRunPrivate *prun = (PangoLayoutRunPrivate *) run;*/
  /*FT_UInt glyph_index;*/
  int i;
  int x_position = 0;
  PangoGPFont *gpfont = ((PangoGPFont *) font);
  PangoGlyphInfo *gi;
  GSList *extra_attrs_list = run->item->analysis.extra_attrs;
  gboolean foreground_set = FALSE;
  guint16 foreground_red = 0, foreground_green = 0, foreground_blue = 0;
  gboolean background_set = FALSE;
  guint16 background_red = 0, background_green = 0, background_blue = 0;
  gint rise = 0;
  gint baseline_adjust = 0;
  GnomePrintContext *gpc;
  GnomeFont *gnome_font = NULL;
  GnomeGlyphList *glyph_list = NULL;

  g_return_if_fail (PANGO_IS_GP_CONTEXT (context));
  g_return_if_fail (run != NULL);
  g_return_if_fail (glyphs != NULL);

  gpc = pango_gp_context_get_context (context);

  glyph_list = gnome_glyphlist_new ();

  gnome_print_moveto (gpc,
		      x / PANGO_SCALE,
		      y / PANGO_SCALE);

  while (extra_attrs_list)
    {
      PangoAttribute *attr = extra_attrs_list->data;
      PangoAttrType attr_type = attr->klass->type;

      switch (attr_type)
	{
	case PANGO_ATTR_RISE:
	  rise = ((PangoAttrInt *) attr)->value;
	  break;
	case PANGO_ATTR_FOREGROUND:
	  foreground_set = TRUE;
	  foreground_red = ((PangoAttrColor *) attr)->color.red;
	  foreground_green = ((PangoAttrColor *) attr)->color.green;
	  foreground_blue = ((PangoAttrColor *) attr)->color.blue;
	  break;
	case PANGO_ATTR_BACKGROUND:
	  background_set = TRUE;
	  background_red = ((PangoAttrColor *) attr)->color.red;
	  background_green = ((PangoAttrColor *) attr)->color.green;
	  background_blue = ((PangoAttrColor *) attr)->color.blue;
	  break;
	default:
	  if (attr_type == pango_gp_attr_callback_get_type ())
	    {
	      GValue values[1] = { { 0, } };

	      g_value_init (&values[0], G_TYPE_POINTER);
	      g_value_set_pointer (&values[0], gpc);

#if 0
	      g_message ("Got a callback: callback: %p",
			 ((PangoGPAttrPointer *) attr)->pointer);
#endif

	      gnome_print_translate (gpc,
				     ((float) x) / PANGO_SCALE,
				     ((float) y) / PANGO_SCALE);

	      g_closure_invoke (((GClosure *) ((PangoGPAttrPointer *) attr)->pointer),
				NULL,
				1,
				values,
				NULL);
	      gnome_print_translate (gpc,
				     - (((float) x) / PANGO_SCALE),
				     - (((float) y) / PANGO_SCALE));
	    }
	  break;
	}
      extra_attrs_list = extra_attrs_list->next;
    }

  if (background_set)
    {
      PangoRectangle logical;

      pango_glyph_string_extents (glyphs,
				  font,
				  NULL,
				  &logical);
      gnome_print_setrgbcolor (gpc,
			       (float) background_red / 0xFFFF,
			       (float) background_green / 0xFFFF,
			       (float) background_blue / 0xFFFF);


      gnome_print_rect_filled (gpc,
			       ((float) (x + logical.x)) / PANGO_SCALE,
			       ((float) y + rise + baseline_adjust - logical.y) / PANGO_SCALE,
			       ((float) logical.width) / PANGO_SCALE,
			       ((float) -logical.height) / PANGO_SCALE);
    }

  if (foreground_set)
    {
      gnome_glyphlist_color (glyph_list,
			     (foreground_red << 24) |
			     (foreground_green << 16) |
			     (foreground_blue << 8));
    }

  gnome_font = gpfont->gnome_font;

  if (gnome_font == NULL)
    {
      if (gpfont->description == NULL)
	{
	  gpfont->description =
	    _pango_gp_font_desc_from_pattern (gpfont->font_pattern, TRUE);
	}

      if (gpfont->gnome_face == NULL)
	{
	  gpfont->gnome_face =
	    gnome_font_face_find_closest_from_pango_description (gpfont->description);
	}

      /*
       * libgnomeprint-2.2.1.2 gnome-font-face.c ignores xres & yres
       * arguments to gnome_font_face_get_font().
       */
      gnome_font = gpfont->gnome_font =
	gnome_font_face_get_font (gpfont->gnome_face,
				  gpfont->size / PANGO_SCALE,
				  72 * PANGO_SCALE,
				  72 * PANGO_SCALE);
    }

  if (gnome_font == NULL)
    {
      g_warning ("No font file for font.");
      return;
    }

  gnome_glyphlist_font (glyph_list,
			gnome_font);

  gi = glyphs->glyphs;
  for (i = 0; i < glyphs->num_glyphs; i++, gi++)
    {
      if (gi->glyph)
	{
	  gnome_glyphlist_glyph (glyph_list,
				 gi->glyph);
	}

      x_position += glyphs->glyphs[i].geometry.width;
    }

  gnome_print_glyphlist (gpc, glyph_list);

  gnome_glyphlist_unref (glyph_list);
}

static FT_Glyph_Metrics *
pango_gp_get_per_char (PangoFont *font,
			guint32    glyph_index)
{
  PangoGPFont *gpfont = (PangoGPFont *)font;
  FT_Face face;
	
  face = pango_gp_font_get_face (font);
  
  FT_Load_Glyph (face, glyph_index, gpfont->load_flags);
  return &face->glyph->metrics;
}

static PangoGPGlyphInfo *
pango_gp_font_get_glyph_info (PangoFont   *font,
			       PangoGlyph   glyph,
			       gboolean     create)
{
  PangoGPFont *gpfont = (PangoGPFont *)font;
  PangoGPGlyphInfo *info;
  FT_Glyph_Metrics *gm;

  info = g_hash_table_lookup (gpfont->glyph_info, GUINT_TO_POINTER (glyph));

  if ((info == NULL) && create)
    {
      FT_Face face = pango_gp_font_get_face (font);
      info = g_new0 (PangoGPGlyphInfo, 1);
      
      if (glyph && (gm = pango_gp_get_per_char (font, glyph)))
	{
#if 0
	  info->ink_rect.x = PANGO_UNITS_26_6_PX (gm->horiBearingX);
	  info->ink_rect.width = PANGO_UNITS_26_6_PX (gm->width);
	  info->ink_rect.y = -PANGO_UNITS_26_6_PX (gm->horiBearingY);
	  info->ink_rect.height = PANGO_UNITS_26_6_PX (gm->height);
	      
	  info->logical_rect.x = 0;
	  info->logical_rect.width = PANGO_UNITS_26_6_PX (gm->horiAdvance);
	  info->logical_rect.y = -PANGO_UNITS_26_6_PX (face->size->metrics.ascender + 64);
	  /* Some fonts report negative descender, some positive ! (?) */
	  info->logical_rect.height = PANGO_UNITS_26_6_PX (face->size->metrics.ascender + ABS (face->size->metrics.descender) + 128);
#endif
	  info->ink_rect.x = gm->horiBearingX;
	  info->ink_rect.width = gm->width;
	  info->ink_rect.y = -gm->horiBearingY;
	  info->ink_rect.height = gm->height;
	      
	  info->logical_rect.x = 0;
	  info->logical_rect.width = gm->horiAdvance;
	  info->logical_rect.y = -face->size->metrics.ascender;
	  /* Some fonts report negative descender, some positive ! (?) */
	  info->logical_rect.height = face->size->metrics.ascender + ABS (face->size->metrics.descender);
	}
      else
	{
	  info->ink_rect.x = 0;
	  info->ink_rect.width = 0;
	  info->ink_rect.y = 0;
	  info->ink_rect.height = 0;

	  info->logical_rect.x = 0;
	  info->logical_rect.width = 0;
	  info->logical_rect.y = 0;
	  info->logical_rect.height = 0;
	}

      g_hash_table_insert (gpfont->glyph_info, GUINT_TO_POINTER(glyph), info);
    }

  return info;
}
     
static void
pango_gp_font_get_glyph_extents (PangoFont      *font,
				  PangoGlyph      glyph,
				  PangoRectangle *ink_rect,
				  PangoRectangle *logical_rect)
{
  PangoGPGlyphInfo *info;

  info = pango_gp_font_get_glyph_info (font, glyph, TRUE);

  if (ink_rect)
    *ink_rect = info->ink_rect;
  if (logical_rect)
    *logical_rect = info->logical_rect;
}

/**
 * pango_gp_font_get_kerning:
 * @font: a #PangoFont
 * @left: the left #PangoGlyph
 * @right: the right #PangoGlyph
 * 
 * Retrieves kerning information for a combination of two glyphs.
 * 
 * Return value: The amount of kerning (in Pango units) to apply for 
 * the given combination of glyphs.
 **/
int
pango_gp_font_get_kerning (PangoFont *font,
			    PangoGlyph left,
			    PangoGlyph right)
{
  FT_Face face;
  FT_Error error;
  FT_Vector kerning;

  face = pango_gp_font_get_face (font);
  if (!face)
    return 0;

  if (!FT_HAS_KERNING (face))
    return 0;

  if (!left || !right)
    return 0;

  error = FT_Get_Kerning (face, left, right,
			  ft_kerning_default, &kerning);
  if (error != FT_Err_Ok)
    g_warning ("FT_Get_Kerning returns error: %s",
	       _pango_gp_ft_strerror (error));

#if 0
  return PANGO_UNITS_26_6_PX (kerning.x);
#endif
  return kerning.x;
}

static PangoFontMetrics *
pango_gp_font_get_metrics (PangoFont     *font,
			    PangoLanguage *language)
{
  PangoGPFont *gpfont = PANGO_GP_FONT (font);
  PangoGPMetricsInfo *info = NULL; /* Quiet gcc */
  GSList *tmp_list;      

  const char *sample_str = pango_language_get_sample_string (language);
  
  tmp_list = gpfont->metrics_by_lang;
  while (tmp_list)
    {
      info = tmp_list->data;
      
      if (info->sample_str == sample_str)    /* We _don't_ need strcmp */
	break;

      tmp_list = tmp_list->next;
    }

  if (!tmp_list)
    {
      PangoContext *context;
      PangoLayout *layout;
      PangoRectangle extents;
      FT_Face face = pango_gp_font_get_face (font);

      info = g_new (PangoGPMetricsInfo, 1);
      info->sample_str = sample_str;
      info->metrics = pango_font_metrics_new ();

#if 0
      info->metrics->ascent = PANGO_UNITS_26_6 (face->size->metrics.ascender);
      info->metrics->descent = PANGO_UNITS_26_6 (- face->size->metrics.descender);
      info->metrics->approximate_char_width = 
        info->metrics->approximate_digit_width = 
        PANGO_UNITS_26_6 (face->size->metrics.max_advance);
#endif
      info->metrics->ascent = face->size->metrics.ascender;
      info->metrics->descent = - face->size->metrics.descender;
      info->metrics->approximate_char_width = 
        info->metrics->approximate_digit_width = 
        face->size->metrics.max_advance;

      gpfont->metrics_by_lang = g_slist_prepend (gpfont->metrics_by_lang, info);

      context = pango_context_new ();
      pango_context_set_font_map (context, gpfont->fontmap);
      pango_context_set_language (context, language);
      
      layout = pango_layout_new (context);
      pango_layout_set_font_description (layout, gpfont->description);

      pango_layout_set_text (layout, sample_str, -1);      
      pango_layout_get_extents (layout, NULL, &extents);
      
      info->metrics->approximate_char_width = 
        extents.width / g_utf8_strlen (sample_str, -1);

      pango_layout_set_text (layout, "0123456789", -1);      
      pango_layout_get_extents (layout, NULL, &extents);
      
      info->metrics->approximate_digit_width = extents.width / 10;

      g_object_unref (G_OBJECT (layout));
      g_object_unref (G_OBJECT (context));
    }

  return pango_font_metrics_ref (info->metrics);
}

static void
pango_gp_font_dispose (GObject *object)
{
  PangoGPFont *gpfont = PANGO_GP_FONT (object);

  /* If the font is not already in the freed-fonts cache, add it,
   * if it is already there, do nothing and the font will be
   * freed.
   */
  if (!gpfont->in_cache && gpfont->fontmap)
    _pango_gp_font_map_cache_add (gpfont->fontmap, gpfont);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}


static gboolean
pango_gp_free_glyph_info_callback (gpointer key, gpointer value, gpointer data)
{
  PangoGPFont *font = PANGO_GP_FONT (data);
  PangoGPGlyphInfo *info = value;
  
  if (font->glyph_cache_destroy && info->cached_glyph)
    (*font->glyph_cache_destroy) (info->cached_glyph);

  g_free (value);
  return TRUE;
}

static void
free_metrics_info (PangoGPMetricsInfo *info)
{
  pango_font_metrics_unref (info->metrics);
  g_free (info);
}

static void
pango_gp_font_finalize (GObject *object)
{
  PangoGPFont *gpfont = (PangoGPFont *)object;

  _pango_gp_font_map_remove (gpfont->fontmap, gpfont);

  if (gpfont->face)
    {
      FT_Done_Face (gpfont->face);
      gpfont->face = NULL;
    }

  pango_font_description_free (gpfont->description);
  FcPatternDestroy (gpfont->font_pattern);
  
  g_object_unref (G_OBJECT (gpfont->fontmap));

  g_free (gpfont->postscript_name);

  g_slist_foreach (gpfont->metrics_by_lang, (GFunc)free_metrics_info, NULL);
  g_slist_free (gpfont->metrics_by_lang);  

  g_hash_table_foreach_remove (gpfont->glyph_info,
			       pango_gp_free_glyph_info_callback, object);
  g_hash_table_destroy (gpfont->glyph_info);
  
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static PangoFontDescription *
pango_gp_font_describe (PangoFont *font)
{
  PangoGPFont         *gpfont;
  PangoFontDescription *desc;

  gpfont = PANGO_GP_FONT (font);

  desc = pango_font_description_copy (gpfont->description);

  return desc;
}

PangoMap *
pango_gp_get_shaper_map (PangoLanguage *language)
{
  static guint engine_type_id = 0;
  static guint render_type_id = 0;
  
  if (engine_type_id == 0)
    {
      engine_type_id = g_quark_from_static_string (PANGO_ENGINE_TYPE_SHAPE);
      render_type_id = g_quark_from_static_string (PANGO_RENDER_TYPE_GP);
    }
  
  return pango_find_map (language, engine_type_id, render_type_id);
}

/**
 * pango_gp_font_get_coverage:
 * @font: a #PangoGPFont.
 * @language: a language tag.
 * @returns: a #PangoCoverage.
 * 
 * Should not be called directly, use pango_font_get_coverage() instead.
 **/
PangoCoverage *
pango_gp_font_get_coverage (PangoFont     *font,
			     PangoLanguage *language)
{
  PangoGPFont *gpfont = (PangoGPFont *)font;

  return _pango_gp_font_map_get_coverage (gpfont->fontmap, gpfont->font_pattern);
}

static PangoEngineShape *
pango_gp_font_find_shaper (PangoFont     *font,
			    PangoLanguage *language,
			    guint32        ch)
{
  PangoMap *shape_map = NULL;

  shape_map = pango_gp_get_shaper_map (language);
  return (PangoEngineShape *)pango_map_get_engine (shape_map, ch);
}

/* Utility functions */

/**
 * pango_gp_get_unknown_glyph:
 * @font: a #PangoFont
 * 
 * Return the index of a glyph suitable for drawing unknown characters.
 * 
 * Return value: a glyph index into @font
 **/
PangoGlyph
pango_gp_get_unknown_glyph (PangoFont *font)
{
  return 0;
}


/**
 * pango_gp_render_layout_line:
 * @context:     the #PangoContext representing where to render.
 * @layout_text: the text of the layout.
 * @line:        a #PangoLayoutLine
 * @x:           the x position of start of string (in #PangoGlyphUnit)
 * @y:           the y position of baseline (in #PangoGlyphUnit)
 *
 * Render a #PangoLayoutLine onto a FreeType2 bitmap
 */
void 
pango_gp_render_layout_line (PangoContext *context,
			     const gchar  *layout_text,
			     PangoLayoutLine *line,
			     gint             x, 
			     gint             y)
{
  GSList *run_list;
  PangoRectangle overall_rect;
  PangoRectangle logical_rect;
  gint      x_off = 0;

  pango_layout_line_get_extents (line,
				 NULL,
				 &overall_rect);
  
  run_list = line->runs;
  while (run_list)
    {
      PangoLayoutRun *run = run_list->data;
      
      pango_glyph_string_extents (run->glyphs,
				  run->item->analysis.font,
				  NULL,
				  &logical_rect);

      pango_gp_render (context,
		       layout_text,
		       run,
		       run->item->analysis.font,
		       run->glyphs,
		       x + x_off,
		       y);

      x_off += logical_rect.width;
      run_list = run_list->next;
    }
}

/**
 * pango_gp_render_layout_lines:
 * @context:     the #PangoContext representing where to render.
 * @layout:      a #PangoLayout
 * @line_first:  the first line to render.
 * @line_last:   the last line to render.
 * @x:           the X position of the left of the layout (in #PangoGlyphUnit)
 * @y:           the Y position of the top of the layout (in #PangoGlyphUnit)
 *
 * Render some lines from #PangoLayout onto a Gp PDF document.
 */
void 
pango_gp_render_layout_lines (PangoContext *context,
			      PangoLayout *layout,
			      gint         line_first,
			      gint         line_last,
			      gint         x, 
			      gint         y)
{
  PangoRectangle layout_logical_rect;
  PangoAlignment align;
  int indent;
  int width;
  int y_offset = 0;
  gint line_index;
  const gchar *layout_text = pango_layout_get_text (layout);
  
  g_return_if_fail (PANGO_IS_GP_CONTEXT (context));
  g_return_if_fail (PANGO_IS_LAYOUT (layout));
  g_return_if_fail (line_first >= 0);
  g_return_if_fail (line_last >= line_first &&
		    line_last <= g_slist_length (pango_layout_get_lines (layout)) - 1);

  indent = pango_layout_get_indent (layout);
  width = pango_layout_get_width (layout);
  align = pango_layout_get_alignment (layout);

  PING (("x:%d y:%d indent:%d width:%d", x, y, indent, width));

  if (width == -1 && align != PANGO_ALIGN_LEFT)
    {
      pango_layout_get_extents (layout, NULL, &layout_logical_rect);
      width = layout_logical_rect.width;
    }
  
  for (line_index = line_first; line_index <= line_last; line_index++)
    {
      PangoLayoutLine *line;
      PangoRectangle   line_ink_rect;
      PangoRectangle   line_logical_rect;
      int x_offset;
      
      line = pango_layout_get_line (layout, line_index);

      pango_layout_line_get_extents (line,
				     &line_ink_rect,
				     &line_logical_rect);

      if (width != 1 && align == PANGO_ALIGN_RIGHT)
	x_offset = width - line_ink_rect.width;
      else if (width != 1 && align == PANGO_ALIGN_CENTER)
	x_offset = (width - line_ink_rect.width) / 2;
      else
	x_offset = 0;

      if (line_index == 0)
	{
	  if (indent > 0)
	    {
	      if (align == PANGO_ALIGN_LEFT)
		x_offset += indent;
	      else
		x_offset -= indent;
	    }
	}
      else
	{
	  if (indent < 0)
	    {
	      if (align == PANGO_ALIGN_LEFT)
		x_offset -= indent;
	      else
		x_offset += indent;
	    }
	}

      PING (("x_offset:%d y_offset:%d logical_rect.y:%d logical_rect.height:%d", x_offset, y_offset, line_logical_rect.y, line_logical_rect.height));

      pango_gp_render_layout_line (context,
				   layout_text,
				   line,
				   x + x_offset,
				   y + y_offset + line_logical_rect.y);

      y_offset -= line_logical_rect.height;
    }
}

/**
 * pango_gp_render_layout:
 * @context:     the #PangoContext representing where to render.
 * @layout:      a #PangoLayout
 * @x:           the X position of the left of the layout (in #PangoGlyphUnit)
 * @y:           the Y position of the top of the layout (in #PangoGlyphUnit)
 *
 * Render a #PangoLayout onto a Gp PDF document.
 **/
void 
pango_gp_render_layout (PangoContext *context,
			PangoLayout  *layout,
			gint          x, 
			gint          y)
{
  gint line_last;

  g_return_if_fail (PANGO_IS_GP_CONTEXT (context));
  g_return_if_fail (PANGO_IS_LAYOUT (layout));

  line_last =
    ((gint) g_slist_length (pango_layout_get_lines (layout))) - 1;

  pango_gp_render_layout_lines (context,
				layout,
				0,
				line_last,
				x,
				y);
}

#if 0
/* This utility function is duplicated here and in pango-layout.c; should it be
 * public? Trouble is - what is the appropriate set of properties?
 */
static void
pango_gp_get_item_properties (PangoItem      *item,
				  PangoUnderline *uline,
				  PangoAttrColor *fg_color,
				  gboolean       *fg_set,
				  PangoAttrColor *bg_color,
				  gboolean       *bg_set)
{
  GSList *tmp_list = item->analysis.extra_attrs;

  if (fg_set)
    *fg_set = FALSE;
  
  if (bg_set)
    *bg_set = FALSE;
  
  while (tmp_list)
    {
      PangoAttribute *attr = tmp_list->data;

      switch (attr->klass->type)
	{
	case PANGO_ATTR_UNDERLINE:
	  if (uline)
	    *uline = ((PangoAttrInt *)attr)->value;
	  break;
	  
	case PANGO_ATTR_FOREGROUND:
	  if (fg_color)
	    *fg_color = *((PangoAttrColor *)attr);
	  if (fg_set)
	    *fg_set = TRUE;
	  
	  break;
	  
	case PANGO_ATTR_BACKGROUND:
	  if (bg_color)
	    *bg_color = *((PangoAttrColor *)attr);
	  if (bg_set)
	    *bg_set = TRUE;
	  
	  break;
	  
	default:
	  break;
	}
      tmp_list = tmp_list->next;
    }
}
#endif

typedef struct
{
  FT_Error     code;
  const char*  msg;
} ft_error_description;

static int
ft_error_compare (const void *pkey,
		  const void *pbase)
{
  return ((ft_error_description *) pkey)->code - ((ft_error_description *) pbase)->code;
}

const char *
_pango_gp_ft_strerror (FT_Error error)
{
#undef __FTERRORS_H__
#define FT_ERRORDEF( e, v, s )  { e, s },
#define FT_ERROR_START_LIST  {
#define FT_ERROR_END_LIST    { 0, 0 } };

  const ft_error_description ft_errors[] =
#include FT_ERRORS_H

#undef FT_ERRORDEF
#undef FT_ERROR_START_LIST
#undef FT_ERROR_END_LIST

  ft_error_description *found =
    bsearch (&error, ft_errors, G_N_ELEMENTS (ft_errors) - 1,
	     sizeof (ft_errors[0]), ft_error_compare);
  if (found != NULL)
    return found->msg;
  else
    {
      static char default_msg[100];

      sprintf (default_msg, "Unknown FreeType2 error %#x", error);
      return default_msg;
    }
}


void *
pango_gp_font_get_cache_glyph_data (PangoFont *font,
				     int        glyph_index)
{
  PangoGPGlyphInfo *info;

  g_return_val_if_fail (PANGO_GP_IS_FONT (font), NULL);
  
  info = pango_gp_font_get_glyph_info (font, glyph_index, FALSE);

  if (info == NULL)
    return NULL;
  
  return info->cached_glyph;
}

void
pango_gp_font_set_cache_glyph_data (PangoFont     *font,
				     int            glyph_index,
				     void          *cached_glyph)
{
  PangoGPGlyphInfo *info;

  g_return_if_fail (PANGO_GP_IS_FONT (font));
  
  info = pango_gp_font_get_glyph_info (font, glyph_index, TRUE);

  info->cached_glyph = cached_glyph;

  /* TODO: Implement limiting of the number of cached glyphs */
}

void
pango_gp_font_set_glyph_cache_destroy (PangoFont      *font,
					GDestroyNotify  destroy_notify)
{
  g_return_if_fail (PANGO_GP_IS_FONT (font));
  
  PANGO_GP_FONT (font)->glyph_cache_destroy = destroy_notify;
}

/**
 * pango_gp_attr_alignment_adjust_get_type:
 * 
 * Register the 'alignment-adjust' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_alignment_adjust_get_type (void)
{
  static PangoAttrType attr_alignment_adjust = PANGO_ATTR_INVALID;

  if (attr_alignment_adjust == PANGO_ATTR_INVALID)
    {
      attr_alignment_adjust =
	pango_attr_type_register ("alignment-adjust");
    }

  return attr_alignment_adjust;
}

/**
 * pango_gp_attr_alignment_adjust_new:
 * @alignment_adjust: the alignment-adjust value
 * 
 * Create a new alignment-adjust attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_alignment_adjust_new (gint alignment_adjust)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_alignment_adjust_get_type();
      klass.copy = pango_attr_int_copy;
      klass.destroy = pango_attr_int_destroy;
      klass.equal = pango_attr_int_equal;
    }

  return pango_attr_int_new (&klass, alignment_adjust);
}

/**
 * pango_gp_attr_alignment_baseline_get_type:
 * 
 * Register the 'alignment-baseline' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_alignment_baseline_get_type (void)
{
  static PangoAttrType attr_alignment_baseline = PANGO_ATTR_INVALID;

  if (attr_alignment_baseline == PANGO_ATTR_INVALID)
    {
      attr_alignment_baseline =
	pango_attr_type_register ("alignment-baseline");
    }

  return attr_alignment_baseline;
}

/**
 * pango_gp_attr_alignment_baseline_new:
 * @alignment_baseline: the alignment-baseline value
 * 
 * Create a new alignment-baseline attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_alignment_baseline_new (gint alignment_baseline)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_alignment_baseline_get_type();
      klass.copy = pango_attr_int_copy;
      klass.destroy = pango_attr_int_destroy;
      klass.equal = pango_attr_int_equal;
    }

  return pango_attr_int_new (&klass, alignment_baseline);
}

/**
 * pango_gp_attr_db_reset_size_get_type:
 * 
 * Register the 'dominant-baseline="reset-size"' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_db_reset_size_get_type (void)
{
  static PangoAttrType attr_db_reset_size = PANGO_ATTR_INVALID;

  if (attr_db_reset_size == PANGO_ATTR_INVALID)
    {
      attr_db_reset_size =
	pango_attr_type_register ("dominant-baseline=\"reset-size\"");
    }

  return attr_db_reset_size;
}

/**
 * pango_gp_attr_db_reset_size_new:
 * @db_reset_size: the dominant baseline.
 * 
 * Create a new dominant-baseline attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_db_reset_size_new (gboolean db_reset_size)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_db_reset_size_get_type();
      klass.copy = pango_attr_int_copy;
      klass.destroy = pango_attr_int_destroy;
      klass.equal = pango_attr_int_equal;
    }

  return pango_attr_int_new (&klass, (int) db_reset_size);
}

/**
 * pango_gp_attr_dominant_baseline_get_type:
 * 
 * Register the '' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_dominant_baseline_get_type (void)
{
  static PangoAttrType attr_dominant_baseline = PANGO_ATTR_INVALID;

  if (attr_dominant_baseline == PANGO_ATTR_INVALID)
    {
      attr_dominant_baseline =
	pango_attr_type_register ("dominant-baseline");
    }

  return attr_dominant_baseline;
}

static PangoAttribute *
pango_gp_attr_baseline_copy (const PangoAttribute *attr)
{
  const PangoGPAttrBaseline *baseline_attr = (PangoGPAttrBaseline *)attr;
  
  return pango_gp_attr_dominant_baseline_new (baseline_attr->baseline, baseline_attr->desc);
}

static void
pango_gp_attr_baseline_destroy (PangoAttribute *attr)
{
  PangoGPAttrBaseline *baseline = (PangoGPAttrBaseline *) attr;

  pango_font_description_free (baseline->desc);
  g_free (attr);
}

static gboolean
pango_gp_attr_baseline_equal (const PangoAttribute *attr1,
			const PangoAttribute *attr2)
{
  const PangoGPAttrBaseline *baseline_attr1 = (const PangoGPAttrBaseline *)attr1;
  const PangoGPAttrBaseline *baseline_attr2 = (const PangoGPAttrBaseline *)attr2;
  
  return ((baseline_attr1->baseline == baseline_attr2->baseline) &&
	  pango_font_description_equal (baseline_attr1->desc,
					baseline_attr2->desc));
}

/**
 * pango_gp_attr_dominant_baseline_new:
 * @dominant_baseline: the dominant baseline.
 * @font_desc:         the #PangoFontDescription.
 * 
 * Create a new dominant-baseline attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_dominant_baseline_new (PangoGPBaseline dominant_baseline,
					 PangoFontDescription *font_desc)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};
  PangoGPAttrBaseline *result;

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_dominant_baseline_get_type();
      klass.copy = pango_gp_attr_baseline_copy;
      klass.destroy = pango_gp_attr_baseline_destroy;
      klass.equal = pango_gp_attr_baseline_equal;
    }

  result = g_new (PangoGPAttrBaseline, 1);
  result->attr.klass = &klass;
  result->baseline = dominant_baseline;
  result->desc = pango_font_description_copy (font_desc);

  return (PangoAttribute *) result;
}

/**
 * pango_gp_attr_callback_get_type:
 * 
 * Register the 'callback' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_callback_get_type (void)
{
  static PangoAttrType attr_callback = PANGO_ATTR_INVALID;

  if (attr_callback == PANGO_ATTR_INVALID)
    {
      attr_callback =
	pango_attr_type_register ("callback");
    }

  return attr_callback;
}

static PangoAttribute *
pango_gp_attr_pointer_new (const PangoAttrClass *klass,
			       gpointer              pointer)
{
  PangoGPAttrPointer *result = g_new (PangoGPAttrPointer, 1);
  result->attr.klass = klass;
  result->pointer = pointer;

  return (PangoAttribute *)result;
}

static PangoAttribute *
pango_gp_attr_pointer_copy (const PangoAttribute *attr)
{
  const PangoGPAttrPointer *pointer_attr = (PangoGPAttrPointer *)attr;

  return pango_gp_attr_pointer_new (attr->klass,
					pointer_attr->pointer);
}

static void
pango_gp_attr_pointer_destroy (PangoAttribute *attr)
{
  g_free (attr);
}

static gpointer
pango_gp_attr_pointer_get_pointer (const PangoAttribute *attr)
{
  const PangoGPAttrPointer *pointer_attr = (PangoGPAttrPointer *)attr;
  
  return pointer_attr->pointer;
}

static gboolean
pango_gp_attr_pointer_equal (const PangoAttribute *attr1,
				  const PangoAttribute *attr2)
{
  const PangoGPAttrPointer *pointer_attr1 =
    (const PangoGPAttrPointer *)attr1;
  const PangoGPAttrPointer *pointer_attr2 =
    (const PangoGPAttrPointer *)attr2;
  
  return (pointer_attr1->pointer == pointer_attr2->pointer);
}

/**
 * pango_gp_attr_callback_new:
 * @callback: the callback
 * 
 * Create a new callback attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_callback_new (gpointer callback)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_callback_get_type();
      klass.copy = pango_gp_attr_pointer_copy;
      klass.destroy = pango_gp_attr_pointer_destroy;
      klass.equal = pango_gp_attr_pointer_equal;
    }

  return pango_gp_attr_pointer_new (&klass, callback);
}

/**
 * pango_gp_attr_callback_get_callback:
 * @attr: The 'callback' attribute.
 * 
 * Get the callback associated with @attr.
 * 
 * Return value: The callback.
 **/
gpointer
pango_gp_attr_callback_get_callback (const PangoAttribute *attr)
{
  return pango_gp_attr_pointer_get_pointer (attr);
}

/**
 * pango_gp_attr_keep_together_within_line_get_type:
 * 
 * Register the 'keep-together.within-line' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_keep_together_within_line_get_type (void)
{
  static PangoAttrType attr_keep_together_within_line = PANGO_ATTR_INVALID;

  if (attr_keep_together_within_line == PANGO_ATTR_INVALID)
    {
      attr_keep_together_within_line =
	pango_attr_type_register ("keep-together.within-line");
    }

  return attr_keep_together_within_line;
}

/**
 * pango_gp_attr_keep_together_within_line_new
 * @strength: the strength of the keep.
 * 
 * Create a new 'keep-together.within-line' attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_keep_together_within_line_new (gint strength)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_keep_together_within_line_get_type();
      klass.copy = pango_attr_int_copy;
      klass.destroy = pango_attr_int_destroy;
      klass.equal = pango_attr_int_equal;
    }

  return pango_attr_int_new (&klass, strength);
}

/**
 * pango_gp_attr_keep_with_next_within_line_get_type:
 * 
 * Register the 'keep-with-next.within-line' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_keep_with_next_within_line_get_type (void)
{
  static PangoAttrType attr_keep_with_next_within_line = PANGO_ATTR_INVALID;

  if (attr_keep_with_next_within_line == PANGO_ATTR_INVALID)
    {
      attr_keep_with_next_within_line =
	pango_attr_type_register ("keep-with-next.within-line");
    }

  return attr_keep_with_next_within_line;
}

/**
 * pango_gp_attr_keep_with_next_within_line_new
 * @strength: the strength of the keep.
 * 
 * Create a new keep-with_next.within-line attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_keep_with_next_within_line_new (gint strength)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_keep_with_next_within_line_get_type();
      klass.copy = pango_attr_int_copy;
      klass.destroy = pango_attr_int_destroy;
      klass.equal = pango_attr_int_equal;
    }

  return pango_attr_int_new (&klass, strength);
}

/**
 * pango_gp_attr_keep_with_previous_within_line_get_type:
 * 
 * Register the 'keep-with-previous.within-line' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_keep_with_previous_within_line_get_type (void)
{
  static PangoAttrType attr_keep_with_previous_within_line = PANGO_ATTR_INVALID;

  if (attr_keep_with_previous_within_line == PANGO_ATTR_INVALID)
    {
      attr_keep_with_previous_within_line =
	pango_attr_type_register ("keep-with-previous.within-line");
    }

  return attr_keep_with_previous_within_line;
}

/**
 * pango_gp_attr_keep_with_previous_within_line_new
 * @strength: the strength of the keep.
 * 
 * Create a new keep-with_previous.within-line attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_keep_with_previous_within_line_new (gint strength)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_keep_with_previous_within_line_get_type();
      klass.copy = pango_attr_int_copy;
      klass.destroy = pango_attr_int_destroy;
      klass.equal = pango_attr_int_equal;
    }

  return pango_attr_int_new (&klass, strength);
}

/**
 * pango_gp_attr_line_height_get_type:
 * 
 * Register the 'line_height' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_line_height_get_type (void)
{
  static PangoAttrType attr_line_height = PANGO_ATTR_INVALID;

  if (attr_line_height == PANGO_ATTR_INVALID)
    {
      attr_line_height =
	pango_attr_type_register ("line-height");
    }

  return attr_line_height;
}

/**
 * pango_gp_attr_line_height_new:
 * @line_height: the line-height value
 * 
 * Create a new line-height attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_line_height_new (gint line_height)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_line_height_get_type();
      klass.copy = pango_attr_int_copy;
      klass.destroy = pango_attr_int_destroy;
      klass.equal = pango_attr_int_equal;
    }

  return pango_attr_int_new (&klass, line_height);
}

/**
 * pango_gp_attr_line_stacking_strategy_get_type:
 * 
 * Register the 'line-stacking-strategy' Pango attribute type.
 * 
 * Return value: The new #PangoAttrType.
 **/
PangoAttrType
pango_gp_attr_line_stacking_strategy_get_type (void)
{
  static PangoAttrType attr_line_stacking_strategy = PANGO_ATTR_INVALID;

  if (attr_line_stacking_strategy == PANGO_ATTR_INVALID)
    {
      attr_line_stacking_strategy =
	pango_attr_type_register ("line-stacking-strategy");
    }

  return attr_line_stacking_strategy;
}

/**
 * pango_gp_attr_line_stacking_strategy_new:
 * @line_stacking_strategy: the line-stacking_strategy value
 * 
 * Create a new line-stacking_strategy attribute.
 * 
 * Return value: the new #PangoAttribute.
 **/
PangoAttribute *
pango_gp_attr_line_stacking_strategy_new (PangoGPLineStackingStrategy line_stacking_strategy)
{
  static PangoAttrClass klass = {PANGO_ATTR_INVALID};

  if (klass.type == PANGO_ATTR_INVALID)
    {
      klass.type = pango_gp_attr_line_stacking_strategy_get_type();
      klass.copy = pango_attr_int_copy;
      klass.destroy = pango_attr_int_destroy;
      klass.equal = pango_attr_int_equal;
    }

  return pango_attr_int_new (&klass, line_stacking_strategy);
}

PangoGPBaselineTable*
pango_gp_font_baseline_table_get (PangoGPFont *font)
{
  return NULL;
#if 0
  g_return_val_if_fail (font != NULL, NULL);

  if (font->baseline_table == NULL)
    {
      if (font->n_fonts == 1)
	{
	  font->baseline_table = font->oa[0]->baseline_table;
	}
      else
	{
	  PangoGPBaselineTable *bt = g_new (PangoGPBaselineTable, 1);
	  int i;

	  bt->units_per_EM = 1 << 21;

	  bt->alphabetic = G_MAXINT;
	  bt->ideographic = G_MAXINT;
	  bt->hanging = G_MININT;
	  bt->mathematical = 0;
	  bt->central = 0;
	  bt->middle = 0;
	  bt->text_before_edge = G_MININT;
	  bt->text_after_edge = G_MAXINT;

	  for (i = 0; i < font->n_fonts; i++)
	    {
	      PangoGPBaselineTable *font_bt = font->oa[i]->baseline_table;
	      int mult;

	      switch (font_bt->units_per_EM)
		{
		case 1000:
		  mult = 2097;
		  break;
		case 2048:
		  mult = 1 << 10;
		  break;
		default:
		  mult = (1 << 21) / font_bt->units_per_EM;
		}

	      bt->alphabetic = MIN (bt->alphabetic, font_bt->alphabetic * mult);
	      bt->ideographic =
		MIN (bt->ideographic, font_bt->ideographic * mult);
	      bt->hanging =
		MAX (bt->ideographic, font_bt->ideographic * mult);
	      bt->mathematical += font_bt->mathematical * mult;
	      bt->central += font_bt->central * mult;
	      bt->middle += font_bt->middle * mult;
	      bt->text_before_edge =
		MAX (bt->text_before_edge, font_bt->text_before_edge * mult);
	      bt->text_after_edge =
		MIN (bt->text_after_edge, font_bt->text_after_edge * mult);
	    }

	  bt->mathematical /= font->n_fonts;
	  bt->central /= font->n_fonts;
	  bt->middle /= font->n_fonts;

	  font->baseline_table = bt;
	}
    }

  return font->baseline_table;
#endif
}
